### Non Backwards compatibale changes
### accounts_balances (in case of any error returns 'errors' instead of 'error')
### accounts_frontiers (in case of any error returns 'errors' instead of 'error')
### accounts_representatives (in case of any error returns 'errors' instead of 'error')

### Changed Commands
# accounts_balances (added optional 'include_only_confirmed')

COMMANDS = {
    "accounts_balances": {
        "required": ["accounts"],
        "optional": ["include_only_confirmed"]
    }
}



from enum import IntEnum
from nanorpc.versions import nano_to

import importlib


class NodeVersion(IntEnum):
    # version modules need follow strict naming convention
    NANO_TO = -10
    V23_3 = 1  # v23_3.py
    V24_0 = 2  # v24_0.py
    V25_0 = 3  # v25_0.py
    V25_1 = 4  # no changes
    V26_0 = 5  # no changes


def get_commands_for_version(node_version):
    commands = {}

    if node_version == NodeVersion.NANO_TO:
        return nano_to.COMMANDS
    # Apply changes for each version from v23_3 up to the selected version
    for version in NodeVersion:
        if version <= node_version:
            module_name = f'nanorpc.versions.{version.name.lower()}'
            try:
                module = importlib.import_module(module_name)
                version_commands = getattr(module, 'COMMANDS', {})
                commands.update(version_commands)
            except ImportError:
                # If a specific version module doesn't exist, just continue
                continue

    return commands



from asyncio import TimeoutError, sleep as aio_sleep
from .versions.handler import NodeVersion, get_commands_for_version
from aiohttp import ClientSession, ClientConnectorError, BasicAuth


def __convert_value(value):
    """Converts a value to its appropriate string representation."""
    if isinstance(value, bool):
        return str(value).lower()
    elif isinstance(value, int):
        return str(value)
    else:
        return value


def __strip_trailing_underscore(arg):
    """Removes trailing underscore from a string if it exists."""
    # For example 'async' is reserved in Python but also used as parameter in the rpc calls.
    # We pass 'async_' in the python method and convert it back to 'async' for teh rpc call
    return arg[:-1] if arg.endswith('_') else arg


def generate_method(command, required, optional):
    async def method(self, *args, **kwargs):
        payload = {"action": command}
        for arg, value in zip(required, args):
            payload[arg] = __convert_value(value)
        for arg in optional:
            # Handle arguments that may end with an underscore
            payload_arg = __strip_trailing_underscore(arg)
            if arg in kwargs and kwargs[arg] is not None:
                payload[payload_arg] = __convert_value(kwargs[arg])
        return await self.process_payloads([payload])

    # Set the name and the docstring of the new function
    method.__name__ = command
    method.__doc__ = f"Execute the {command} command."

    return method


class MaxRetriesExceededError(Exception):
    """Raised when all retries are exhausted"""
    pass


class NanoRpc:
    RETRY_ON_EXCEPTIONS = (ClientConnectorError, TimeoutError,
                           ConnectionResetError)

    def __init__(self,
                 url,
                 node_version: NodeVersion,
                 max_retries=3,
                 username=None,
                 password=None,
                 auth_key=None,
                 app_name=None,
                 app_email=None):
        self.url = url
        self.username = username
        self.password = password
        self.max_retries = max_retries
        self.auth = BasicAuth(username,
                              password) if username and password else None
        self.headers = self._set_headers(auth_key, app_name, app_email)
        self.session = None
        self.commands = get_commands_for_version(node_version)
        self._generate_methods()

    def _set_headers(self, auth_key, app_name, app_email):
        headers = {"Authorization": f"Bearer {auth_key}"} if auth_key else {}
        if app_name:
            headers["nano-app"] = app_name
        if app_email:
            headers["nano-app-admi"] = app_email
        return headers

    def _generate_methods(self):
        for command, params in self.commands.items():
            method = generate_method(command, params["required"],
                                     params["optional"])
            bound_method = method.__get__(
                self, NanoRpc)  # This binds the method to the instance
            setattr(self, method.__name__, bound_method)

    async def _retry_on_exception(self, coroutine, *args, retries):
        for retry in range(retries):
            try:
                return await coroutine(*args)
            except self.RETRY_ON_EXCEPTIONS as e:
                if retry == retries - 1:  # If this was the last retry
                    raise MaxRetriesExceededError(
                        f"All {self.max_retries - retry} retries exhausted for {coroutine.__name__}. Last error: {str(e)}"
                    ) from None
                await aio_sleep(0.5 * 2**retry)

    async def _request_with_retry(self, coroutine, *args, **kwargs):
        return await self._retry_on_exception(coroutine,
                                              *args,
                                              retries=self.max_retries,
                                              **kwargs)

    async def _request(self, payloads):
        async with ClientSession(auth=self.auth) as session:
            async with session.post(self.url, json=payloads[0], headers=self.headers) as response:
                return await response.json()

    async def process_payloads(self, payloads):
        return await self._request_with_retry(self._request, payloads)



from nanorpc.client_dynamic import NanoRpc, NodeVersion


class NanoToRpcTyped:
    def __init__(self, auth_key, app_name="nanoto_python_lib", app_email=None):
        self.rpc = NanoRpc(url="https://rpc.nano.to",
                           auth_key=auth_key, app_name=app_name, app_email=app_email, node_version=NodeVersion.NANO_TO)

    async def version(self, ):
        return await self.rpc.version()

    async def account_info(self, account, representative=None, weight=None, receivable=None, include_confirmed=None):
        return await self.rpc.account_info(account, representative=representative, weight=weight, receivable=receivable, include_confirmed=include_confirmed)

    async def receivable(self, account, count=None, threshold=None, source=None, include_active=None, min_version=None, sorting=None, include_only_confirmed=None, offset=None, array=None):
        return await self.rpc.receivable(account, count=count, threshold=threshold, source=source, include_active=include_active, min_version=min_version, sorting=sorting, include_only_confirmed=include_only_confirmed, offset=offset, array=array)

    async def account_balance(self, account):
        return await self.rpc.account_balance(account)

    async def account_history(self, account, count=None, raw=None, head=None, offset=None, reverse=None, account_filter=None):
        return await self.rpc.account_history(account, count=count, raw=raw, head=head, offset=offset, reverse=reverse, account_filter=account_filter)

    async def accounts_balances(self, accounts, include_only_confirmed=None):
        return await self.rpc.accounts_balances(accounts, include_only_confirmed=include_only_confirmed)

    async def accounts_receivable(self, accounts, count=None, threshold=None, source=None, include_active=None, sorting=None, include_only_confirmed=None):
        return await self.rpc.accounts_receivable(accounts, count=count, threshold=threshold, source=source, include_active=include_active, sorting=sorting, include_only_confirmed=include_only_confirmed)

    async def block_info(self, hash, json_block=None):
        return await self.rpc.block_info(hash, json_block=json_block)

    async def process(self, block, force=None, subtype=None, json_block=None, async_=None):
        return await self.rpc.process(block, force=force, subtype=subtype, json_block=json_block, async_=async_)

    async def work_generate(self, hash, key=None):
        return await self.rpc.work_generate(hash, key=key)

    async def block_count(self, include_cemented=None):
        return await self.rpc.block_count(include_cemented=include_cemented)

    async def account_key(self, account):
        return await self.rpc.account_key(account)

    async def price(self, currency):
        return await self.rpc.price(currency)

    async def reps(self, ):
        return await self.rpc.reps()

    async def rep_info(self, account):
        return await self.rpc.rep_info(account)

    async def nano_to_raw(self, amount):
        return await self.rpc.nano_to_raw(amount)

    async def raw_to_nano(self, amount):
        return await self.rpc.raw_to_nano(amount)

    async def known(self, ):
        return await self.rpc.known()

    async def get_name(self, name):
        return await self.rpc.get_name(name)

    async def update_name(self, name):
        return await self.rpc.update_name(name)

    async def checkout(self, title, notify, webhook):
        return await self.rpc.checkout(title, notify, webhook)

    async def market_data(self, ):
        return await self.rpc.market_data()

    async def nano_swap(self, amount, from_, to, address, refund_address):
        return await self.rpc.nano_swap(amount, from_, to, address, refund_address)

    async def nano_ai(self, model, prompt):
        return await self.rpc.nano_ai(model, prompt)

    async def cloud_wallet(self, refund_address, vanity, password):
        return await self.rpc.cloud_wallet(refund_address, vanity, password)

    async def nano_email(self, amount, email_address, refund_address, claimed_webhook, refunded_webhook):
        return await self.rpc.nano_email(amount, email_address, refund_address, claimed_webhook, refunded_webhook)

    async def buy_rpc(self, email):
        return await self.rpc.buy_rpc(email)

    async def gpu_key(self, email):
        return await self.rpc.gpu_key(email)
